const path = require('path')
const fs = require('fs')
const gulp = require('gulp')
const ansi = require('gulp-cli/lib/shared/ansi')
const exit = require('gulp-cli/lib/shared/exit')
const formatError = require('gulp-cli/lib/versioned/^4.0.0/format-error')
const prettyTime = require('pretty-hrtime')
const ts = require('gulp-typescript')
const less = require('gulp-less')
const rename = require('gulp-rename')
const banner = require('gulp-banner')
const postcss = require('gulp-postcss')
const autoprefixer = require('autoprefixer')
const convertCssVar = require('gulp-convert-css-var')
const px2rpx = require('./gulp.px2rpx')
const util = require('./util')

const buildDir = util.buildDir
const rootDir = util.rootDir

const doraConfig = fs.existsSync(path.join(buildDir, 'dora.config.js')) ?
  path.join(buildDir, 'dora.config.js') :
  path.join(rootDir, 'dora.config.js')

const tsconfig = fs.existsSync(path.join(buildDir, 'tsconfig.json')) ?
  path.join(buildDir, 'tsconfig.json') :
  path.join(rootDir, 'tsconfig.json')

const tsProject = ts.createProject(tsconfig, {
  typescript: require('typescript')
})

const config = Object.assign({
  entry: ['./src/**/*.ts'],
  outputDir: './miniprogram_dist',
  copyPlugin: {
    entry: [
      './src/**/*.json',
      './src/**/*.wxml',
      './src/**/*.wxss',
      '!./src/**/*.ts'
    ]
  },
  cssPlugin: {
    entry: [
      './src/**/*.less'
    ],
    pxTransform: {
      designWidth: 375
    }
  },
}, fs.existsSync(doraConfig) ? require(doraConfig) : {})

const copy = (paths) => () => (
  gulp
    .src(paths.entry)
    .pipe(gulp.dest(paths.outputDir))
)

const scripts = (paths) => () => (
  gulp
    .src(paths.entry)
    .pipe(tsProject())
    .js
    .pipe(banner(util.banner()))
    .pipe(gulp.dest(paths.outputDir))
)

const styles = (paths) => () => (
  gulp
  .src(paths.entry)
  .pipe(less({
    javascriptEnabled: true
  }))
  .pipe(postcss([ autoprefixer() ]))
  .pipe(px2rpx(paths.pxTransform))
  .pipe(convertCssVar())
  .pipe(rename({ extname: '.wxss' }))
  .pipe(gulp.dest(paths.outputDir))
)

const watch = () => {
  gulp.watch(config.copyPlugin.entry, copy({ entry: config.copyPlugin.entry, outputDir: config.outputDir }))
  gulp.watch(config.entry, scripts({ entry: config.entry, outputDir: config.outputDir }))
  gulp.watch(config.cssPlugin.entry, styles({ entry: config.cssPlugin.entry, outputDir: config.outputDir, pxTransform: config.cssPlugin.pxTransform }))
}

gulp.task('copy', copy({ entry: config.copyPlugin.entry, outputDir: config.outputDir }))
gulp.task('scripts', scripts({ entry: config.entry, outputDir: config.outputDir }))
gulp.task('styles', styles({ entry: config.cssPlugin.entry, outputDir: config.outputDir, pxTransform: config.cssPlugin.pxTransform }))
gulp.task('build', gulp.series(['copy', 'scripts', 'styles']))
gulp.task('watch', watch)


// Wire up logging events
function logEvents (gulpInst) {

  const loggedErrors = []

  gulpInst.on('start', function(evt) {
    /* istanbul ignore next */
    // TODO: batch these
    // so when 5 tasks start at once it only logs one time with all 5
    const level = evt.branch ? 'debug' : 'info'
    console[level]('Starting', '\'' + ansi.cyan(evt.name) + '\'...')
  })

  gulpInst.on('stop', function(evt) {
    const time = prettyTime(evt.duration)
    /* istanbul ignore next */
    const level = evt.branch ? 'debug' : 'info'
    console[level](
      'Finished', '\'' + ansi.cyan(evt.name) + '\'',
      'after', ansi.magenta(time)
    )
  })

  gulpInst.on('error', function(evt) {
    const msg = formatError(evt)
    const time = prettyTime(evt.duration)
    const level = evt.branch ? 'debug' : 'error'
    console[level](
      '\'' + ansi.cyan(evt.name) + '\'',
      ansi.red('errored after'),
      ansi.magenta(time)
    )

    // If we haven't logged this before, log it and add to list
    if (loggedErrors.indexOf(evt.error) === -1) {
      console.error(msg)
      loggedErrors.push(evt.error)
    }
  })
}

function execute (opts = {}) {
  const tasks = opts._
  const toRun = tasks.length ? tasks : ['default']
  logEvents(gulp)
  try {
    console.info('Powered by dora')
    const runMethod = opts.series ? 'series' : 'parallel'
    gulp[runMethod](toRun)(function(err) {
      if (err) {
        exit(1)
      }
    })
  } catch (err) {
    console.error(ansi.red(err.message))
    exit(1)
  }
}

module.exports = execute
